"""Creating and managing auditory stimuli
"""
from types import SimpleNamespace
from os import environ as _env
from reiz._audio.primitives import AudioFile, Hertz, Message
from pathlib import Path
from typing import Dict, NewType, Any
libConf = NewType("libConf", Dict[str, Dict[str, Any]]
                  )  #: Dict[str, Dict[str, Any], a dictionary of types and respective keywords arguments

_defaults = libConf({
    "Message": {
        "start": {"message": "start"}
    },
    "Hertz": {
        "beep": {"duration_in_ms": 1000}
    },
})  #: libConf


def make_library(settings: libConf = _defaults, failraise=False) -> SimpleNamespace:
    """create a library of auditory stimuli from a dictionary of arguments

    args
    ----
    settings:
        a dictionary of types with kwargs appropriate to the respective type
        as dictionary
    failraise:bool
        raise an exception if a key is not recognized. defaults to False

    returns
    -------

    library:
        a library of instances of stimuli which can be adressed in dot.notation
    """
    lib = dict()
    for key in settings.keys():
        for name, args in settings[key].items():
            if key.lower() == "audiofile":
                lib[name] = AudioFile(**args)
            elif key.lower() == "message":
                lib[name] = Message(**args)
            elif key.lower() == "hertz":
                lib[name] = Hertz(**args)
            else:
                if failraise:
                    raise ValueError(f"{key} with {args} can't be processed")
    lib = SimpleNamespace(**lib)
    return lib


def read_folder(path: Path = None) -> SimpleNamespace:
    'create an audio library from path'
    path = Path(path)
    if path is None or not path.exists():
        raise ValueError(f"{path} not found")
    library = dict()
    import os
    for f in os.listdir(path):
        key = os.path.splitext(f)[0]
        key = key.replace("(", "")
        key = key.replace(")", "")
        key = key.replace(" ", "_")
        key = key.replace("-", "_")
        key = key.strip()
        val = AudioFile(os.path.join(path, f))
        library[key] = val
    library = SimpleNamespace(**library)
    return library


if not 'DOC' in _env.keys():
    library = make_library()
else:  # pragma: no cover
    mock = dict()
    for t, item in _defaults.items():
        for k, v in item.items():
            mock[k] = t
    library = SimpleNamespace(**mock)  #: a library of auditory stimuli
    print("Generating sphinx documentation. Skipping visual library")


__doc__ += f"""
During import, a basic library will be created. By default, it contains
the following set of auditory stimuli: {', '.join([str(k) for k in library.__dict__])}. These stimuli can be addressed by dot.notation from :data:`reiz.audio.library`, for example like :data:`reiz.audio.library.beep`


.. automodule:: reiz._audio.primitives
   :members: Message, Hertz, AudioFile

Create libraries of auditory stimuli
....................................

"""



from pytest import fixture
import reiz
from reiz._audio.primitives import Sound
from reiz._audio.tts import Silent_Mixin


@fixture
def msg():
    msg = reiz.audio.Message("test")
    msg.volume = 0
    return msg


@fixture
def silence():

    class Silence(Silent_Mixin, Sound):
        pass
    return Silence()


def test_silence(silence):
    assert silence.duration == 0.1


def test_properties(msg):
    assert msg.volume == 0


def test_play_stop(msg):
    dur = msg.duration
    played_time = msg.play()
    msg.stop()
    assert played_time < dur


def test_play_blocking(msg):
    played_time = msg.play_blocking()
    assert played_time == 0



'''Root module for the pyreiz stimulus presentation toolbox
'''


from sys import platform
import pyglet
import reiz.visual as visual
import reiz.audio as audio
import reiz.marker as marker
from reiz.cue import Cue
from reiz._visual._screen import Canvas
from reiz.time import clock, Clock


if platform == 'linux':
    pyglet.options['audio'] = ('openal', 'pulse', 'directsound', 'silent')

del pyglet, platform



# -*- coding: utf-8 -*-
"""
Library of colors
.................
"""
from typing import NewType, Union, Tuple, List

RGB = NewType("RGB", Tuple[float, float, float]
              )  #: Tuple[float, float, float] limited between 0 and 1
ColorType = NewType("ColorType", Union[RGB, str]
                    )  #: Union[RGB, str] either a RGB color or a descriptive string from :data:`~.COLORS`

RGBA = NewType("RGBA", Tuple[float, float, float, float]
               )  #: Tuple[float, float, float, float] limited between 0 and 1
# %%
COLORS = {
    "red": [0.9, 0.1, 0.1],
    'carmine': [.59, .0, .09],
    'pictorial_carmine': (.76, .04, .31),
    'crimson': (.87, .08, .24),
    "pink": [0.9, 0.5, 0.9],
    "green": [.25, .8, .25],
    'moss': (.54, .60, .36),
    'darkgreen': (.0, .39, .0),
    "blue": [0, 0, 1],
    "white": [1, 1, 1],
    "bright": [0.75, 0.75, 0.75],
    "light": [0.5, 0.5, 0.5],
    "gray": [0.25, 0.25, 0.25],
    "dark": [0.125, 0.125, 0.125],
    "darker": [0.1, 0.1, 0.1],
    "black": [0, 0, 0],
    "brown": [.55, .36, .24],
    'darkbrown': [.44, .29, .20],
    'turkis': (.28, .82, .8),
    'darkturkis': (0., .8, .82),
    'celeste': (.69, 1, 1),
}  #: A dictionary of color strings encoding a tuple in RGB


def resolve_rgb(a: RGBA, b: RGBA, n: int = 100) -> List[RGBA, ]:
    '''interpolate an RGBA color in steps

    args
    ----

    a: RGBA
      rgba tuple
    b: RGBA
      rgba tuple
    n: int
      number of steps

    returns
    -------

    color: List[RGBA,]
        list of interpolated rgba tuples

    '''

    def linspace(a, b, n):
        if n < 2:
            return b
        diff = (float(b) - a)/(n - 1)
        return [diff * i + a for i in range(n)]
    if type(a) is str:
        a = get_color(a)
    if type(b) is str:
        b = get_color(b)

    _colors = []
    for _a, _b in zip(a, b):
        _colors.append(linspace(_a, _b, n))
    colors = []
    for i in range(0, n, 1):
        colors.append((_colors[0][i], _colors[1][i], _colors[2][i]))

    return colors


def get_color(color: ColorType = "white", opacity: float = 1) -> RGBA:
    '''get rgba tuple based on color name

    args
    ----
    color: ColorType, i.e. Union[Tuple[float, float, float], str]
        a color name or a tuple of colors, as from :data:`reiz.visual.colors.COLORS`
    opacity: float
        the alpha value


    returns
    -------
    color: RGBA i.e Tuple[float, float, float, float]
        a rgba tuple with values between 0 and 1
    '''

    try:
        color = COLORS[color]
    except KeyError:
        if len(color) != 3:
            raise ValueError('Not a correct color value')
    try:
        color = color.copy()
        color.append(opacity)
    except AttributeError:
        color = (*color, opacity)

    for val in color:
        if val < 0 or val > 1:
            raise ValueError("The desired color is not within range of 0 to 1")

    return color



import reiz


def test_canvas_keypress():
    from pyglet.window import key
    canvas = reiz.Canvas()
    canvas.open()
    assert canvas.start_run == False
    canvas.window.dispatch_event('on_key_press', key.F5, 0)
    canvas.flip()
    assert canvas.start_run == True
    canvas.close()


def test_canvas_available():
    canvas = reiz.Canvas(size=(5, 5))
    assert canvas.available == False
    canvas.open()
    assert canvas.available == True
    canvas.close()
    assert canvas.available == False


def test_canvas_properties():
    w, h = (10, 9)
    canvas = reiz.Canvas(size=(w, h))
    canvas.open()
    canvas.clear()
    canvas.get_diag()
    canvas.estimate_fps()
    assert h == canvas.height
    assert w == canvas.width
    canvas.close()


def test_clock_sleep():
    limit = 1.5 * 10**(-4)  # exact within a millisecond
    desired = .1
    actual = reiz.clock.sleep(desired)
    deviance = abs(desired-actual)
    assert deviance < limit


def test_clock_sleep_debiased():
    import random
    import time
    limit = 1.5 * 10**(-4)  # exact within a millisecond
    desired = .1
    reiz.clock.reset()
    for i in range(1, 11):
        time.sleep(random.random()*0.05)
        reiz.clock.sleep_debiased(desired)
    actual = reiz.clock.now()
    deviance = abs(desired*i-actual)
    assert deviance < limit



"""Sending markers

Sending markers requires a MarkerServer running. This markerserver is used as a man-in-the-middle to forward all markers into an already existing LSLOutlet. 
This MarkerServer opens an Outlet that can be detected independently from the experiments you are running. When you then run an experiment, it receives messages from this experiment, and redistributes them in LSL-format.

LabRecorder can only start recording Outlets that are existing. If you start a MarkerServer immediatly before you run the experiment, there might not be sufficient time for LabRecorder to detect the stream. Certainly, it adds some precious seconds for starting the Outlet, and delays your run.

We recommend to use a reiz-marker-Server as an independent process. Using reiz-marker allows easier recording, reduces computational load on your main script, and can prevent forgetting to record a MarkerStream with LabRecorder. It does though add a little bit more complexity and a miniscule delay.

From the terminal 
-----------------

You can start the MarkerServer as an independent process with `reiz-marker` from your terminal and shut it down gracefull with `reiz-marker --kill`.

.. code-block:: bash

    usage: reiz-marker [-h] [--port PORT] [--host HOST] [--name NAME] [--ping] [--kill]

    Reiz Marker Server

    optional arguments:
    -h, --help   show this help message and exit
    --port PORT  Marker Server port.
    --host HOST  Marker Server host ip.
    --name NAME  Marker Server name.
    --ping       test connection to Markerserver
    --kill       send a poison pill to the Markerserver


From within Python
------------------

Alternatively, :meth:`~.reiz._marker.safeguard.start` starts such a process from within Python, and you can kill it later with :meth:`~.reiz._marker.safeguard.stop`. You can additionally safeguard your script to make sure a MarkerServer is running in at least these two ways:

a) check whether the MarkerServer is avaible, and abort with a notice, if that is not the case.
b) start a throw-away MarkerServer with :meth:`~.start`. 

Sending markers
---------------

There is not much need to push markers for stimuli events manually, as you can create a :class:`~.Cue` and add the desired marker-string during instantiation. Often, you want to send additional information, e.g. the age and id of the subject, or the parameters with which the experiment was run. In these cases, pushing dictionaries or arbitrary strings during initialization of your experiment can be useful.  

.. currentmodule:: reiz._marker.client
.. autosummary::
   :template: module.rst

    push
    push_json



Safeguarding
------------
.. currentmodule:: reiz._marker.safeguard
.. autosummary::
   :template: module.rst

    available
    start
    stop

"""

from reiz._marker.client import push, push_json, available
from reiz._marker.safeguard import start, stop



"""
Safeguard your scripts against absence of a MarkerServer
........................................................


"""
from reiz._marker.mitm import Server
from sys import platform
from reiz._marker.client import _Client
from subprocess import Popen, run
from time import sleep
from sys import platform
from pylsl import local_clock
from logging import getLogger
logger = getLogger("throw-away-marker-server")

server = None


def available(port: int = 7654, verbose=True) -> bool:
    """test whether a markerserver is already available at port

    Example
    -------

    Protect your script by failing when no marker-server is running.
    .. codeblock:: Python

        if not available():
            print("Start independent markerserver with `reiz-marker`")
            quit()

    args
    ----

    port: int
        the port number of the markerserver (defaults to 7654)

    returns
    -------

    status: bool
        True if available, False if not

    """
    host: str = "127.0.0.1"
    c = _Client(host=host, port=port)
    try:
        c.push('None', local_clock())
        return True
    except ConnectionRefusedError as e:
        if verbose:
            print(e)
            print(f'Markerserver at {host}:{port} is not available')
        return False


def start():
    """start a throw-away MarkerServer

    Close it after the experiment with :func:`~.stop`

    .. caution::
        consider that it is best practice to start an independent MarkerServer to
        improve discoverability and stability for any listeners, especially
        when recording with LabRecorder. Use these functions only to safeguard
        your experiment and to be able to test it without the need for an
        unrelated process running.

    """
    global server
    if server is None:
        logger.debug("Starting a marker-server")
        server = Popen(["reiz-marker", "--name", "reiz-marker"])
        while not available(verbose=False):
            sleep(0.5)
        return server
    else:
        logger.debug("A marker-server is already running")
        return None


def stop():
    """stop the throw-away MarkerServer

    This MarkerServer was started earlier with :func:`~.start`
    """
    global server
    if server is None:
        logger.debug("No marker-server is currently running")
        return True
    else:
        killer = Popen(["reiz-marker", "--kill", "--name", "reiz-marker"])
        killer.communicate()
        while available(verbose=False):
            sleep(0.5)
        return True



import setuptools
from distutils.core import setup

# read the contents of README.md
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()
long_description.replace(
    "basic-example.gif",
    "https://raw.githubusercontent.com/pyreiz/pyreiz/master/basic-example.gif",
)


setup(
    name="Reiz",
    version="v0.3.4.1",
    description="A Python toolbox for visual and auditory stimulation based on pyglet and pylsl.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Robert Guggenberger",
    author_email="robert.guggenberger@uni-tuebingen.de",
    url="https://github.com/pyreiz/pyreiz",
    download_url="https://github.com/pyreiz/pyreiz.git",
    license="MIT",
    include_package_data=True,
    package_data={"reiz": ["data/*.*"]},
    packages=setuptools.find_packages(),
    install_requires=["pyglet >= 1.4.7", "pylsl >= 1.13",],
    extras_require={"tts": ["pyttsx3 >= 2.7"]},
    entry_points={"console_scripts": ["reiz-marker=reiz._marker.__main__:main"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Multimedia :: Graphics :: Presentation",
        "Topic :: Multimedia :: Sound/Audio :: Players",
        "Topic :: Scientific/Engineering :: Human Machine Interfaces",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
    ],
)



"""Client interface to the MarkerServer to send markers"""
import pylsl
import socket
import json
from logging import getLogger
log = getLogger()


def sanitize_string(marker: str) -> str:
    """sanitize a string

    removes whitespace, transforms to lower case and replaces umlaute and
    spaces with safer symbols

    args
    ----
    marker:str
        the un-sanitized string

    returns
    -------
    sanitized:str
        the sanitized string
    """
    translation = str.maketrans({'ä': 'ae', 'ö': 'oe', 'ü': 'ue', ' ': '_'})
    marker = marker.lower().strip().translate(translation)

    if marker.lower() == "ping":
        marker = json.dumps({"msg": marker})
        log.critical(
            f"'ping' is a reserved key-word for pinging the marker-server. If you want to ping, use available. Forwarded {marker} instead")
    if marker.lower() == "poison-pill":
        marker = {"msg": marker}
        marker = json.dumps(marker)
        log.critical(
            f"'poison-pill' is a reserved key-word for shutting down the marker-server. If you want to shut-down, use kill. Forwarded {marker} instead")
    return marker


def push(marker: str = '', tstamp: float = None,
         sanitize=True, port: int = 7654):
    """push a marker to the MarkerServer for redistribution as LSL

    args
    ----

    marker: str
        an ascii-encodable string describing an event. We recommend
        to use valid json-strings :func:`~.push_json`
    tstamp: float
        the timestamp of the event. We recommend to use timestamps received
        from pylsl.local_clock
    sanitize: bool
        whether the string is to be sanitized, see :func:`~.sanitize_string`
    port: int
        the port of the MarkerServer

    """
    if tstamp is None:
        tstamp = pylsl.local_clock()
    if sanitize:
        marker = sanitize_string(marker)

    c = _Client(port=port)
    c.push(marker, tstamp)


def push_json(marker: dict = {'key': 'value'}, tstamp: float = None):
    """encode a dictionary as json and push it to the MarkerServer

    args
    ----
    marker:
        a json-encodable dictionary describing an event or current settings
    tstamp:
        the timestamp of the event. We recommend to use timestamps received
        from pylsl.local_clock

    .. caution:
        the resulting msg will not be sanitized :func:`~.sanitize_string`
    """
    push(json.dumps(marker), tstamp, sanitize=False)


def available(port: int = 7654, host: str = "127.0.0.1", verbose=True) -> bool:
    """test whether a markerserver is already available at port

    args
    ----

    host: str
        the ip of the markerserver (defaults to localhost)

    port: int
        the port number of the markerserver (defaults to 7654)

    returns
    -------

    status: bool
        True if available, False if not
    """
    c = _Client(host=host, port=port)
    try:
        c.push('ping', pylsl.local_clock())
        return True
    except ConnectionRefusedError as e:
        if verbose:
            print(e)
            print(f'Markerserver at {host}:{port} is not available')
        return False


def kill(host: str = "127.0.0.1",
         port: int = 7654):
    c = _Client(port=port, host=host)
    c.push("poison-pill",  pylsl.local_clock())


class _Client():
    "Basic Client communicating with the MarkerServer"

    def __init__(self, host="127.0.0.1", port: int = 7654, verbose=True):
        self.host = host
        self.port = port
        self.verbose = verbose

    def push(self, marker: str = '', tstamp: float = None):
        'connects, sends a message, and close the connection'
        self.connect()
        self.write(marker, tstamp)
        self.close()

    def connect(self):
        'connect wth the remote server'
        self.interface = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.interface.connect((self.host, self.port))
        self.interface.settimeout(1)

    def write(self, marker, tstamp):
        'encode message into ascii and send all bytes'
        msg = json.dumps((marker, tstamp)).encode('ascii')
        if self.verbose:
            print(f'Sending {marker} at {tstamp}')
        self.interface.sendall(msg)

    def close(self):
        'closes the connection'
        self.interface.shutdown(1)
        self.interface.close()
